---
-- AdditionalSettings
--
-- Author: Rockstar
-- Version: 1.0.0.0
-- Date: 21.03.2021
--

AdditionalSettings = {
	LOADING_STATE = {
		NO_LOAD = 0,
		ON_CREATE = 1,
		ON_MAP_LOAD = 2,
		ON_MISSION_START = 3
	},
	SHOW_SECONDS = false
}

if AdditionalSettings.SHOW_SECONDS then
	GameInfoDisplay.SIZE.TIME_BOX[1] = 150
end

local AdditionalSettings_mt = Class(AdditionalSettings)

function AdditionalSettings:new(mission, missionInfo, missionDynamicInfo, baseDirectory, modName, i18n, gui, gameSettings, depthOfFieldManager, mpLoadingScreen, shopConfigScreen, mapManager)
	local self = setmetatable({}, AdditionalSettings_mt)

	addModEventListener(self)
	mission:registerObjectToCallOnMissionStart(self)

	self.currentMission = mission
	self.i18n = i18n
	self.gui = gui
	self.gameSettings = gameSettings
	self.depthOfFieldManager = depthOfFieldManager
	self.mpLoadingScreen = mpLoadingScreen
	self.shopConfigScreen = shopConfigScreen
	self.mapManager = mapManager
	self.baseDirectory = baseDirectory

	self:overwriteFunctions(AdditionalSettings.LOADING_STATE.ON_CREATE)

	self.addittionalSettings = self:createSettings()

	self:loadSettingsFromXMLFile(g_modsSettingsPath, AdditionalSettings.LOADING_STATE.ON_CREATE)
	self:copyModEnvironmentTexts(true)
	self:addPage()

	return self
end

function AdditionalSettings:addPage()
	local additionalSettingsFrame = AdditionalSettingsFrame:new(nil, self.i18n)
	local newFrameReference = NewFrameReference:new(nil)
	local restartGameDialog = RestartGameDialog:new(nil)

	self.gui:loadGui(Utils.getFilename("xml/gui/AdditionalSettingsFrame.xml", self.baseDirectory), "AdditionalSettingsFrame", additionalSettingsFrame, true)
	self.gui:loadGui(Utils.getFilename("xml/gui/NewFrameReference.xml", self.baseDirectory), "NewSettings", newFrameReference)
	self.gui:loadGui(Utils.getFilename("xml/gui/RestartGameDialog.xml", self.baseDirectory), "RestartGameDialog", restartGameDialog)

	local inGameMenu = self.currentMission.inGameMenu
	local pageAdditionalSettings = newFrameReference.pageAdditionalSettings

	if pageAdditionalSettings ~= nil then
		local position = inGameMenu.pagingElement:getPageIndexByElement(inGameMenu.pageSettingsGeneral) + 1

		inGameMenu.pagingElement.addPage = Utils.overwrittenFunction(inGameMenu.pagingElement.addPage, function(object, superFunc, id, element, title, index)
			if element.name == "ingameMenuAdditionalSettings" then
				index = position
			end

			return superFunc(object, id, element, title, index)
		end)

		inGameMenu.pagingElement:addElement(pageAdditionalSettings)
		inGameMenu:registerPage(pageAdditionalSettings, position, inGameMenu:makeIsGeneralSettingsEnabledPredicate())
		inGameMenu:addPageTab(pageAdditionalSettings, g_baseUIFilename, getNormalizedUVs({390, 144, 65, 65}))
		inGameMenu.notifySaveComplete = Utils.appendedFunction(inGameMenu.notifySaveComplete, pageAdditionalSettings.notifySaveComplete)
		inGameMenu.pageAdditionalSettings = pageAdditionalSettings
	end
end

function AdditionalSettings:createSettings()
	local settings = {
		hud = {
			active = true,
			target = self,
			callbackFunc = self.changeHudVisibility,
			updateFunc = self.updateHudVisibility,
			loadingState = AdditionalSettings.LOADING_STATE.ON_MISSION_START,
			name = "hud"
		},
		hourFormat = {
			active = true,
			target = self,
			callbackFunc = self.changeHourFormat,
			loadingState = AdditionalSettings.LOADING_STATE.ON_CREATE,
			name = "hourFormat"
		},
		dialogs = {
			active = true,
			target = self,
			createFunc = self.createDialogsSettings,
			loadingState = AdditionalSettings.LOADING_STATE.ON_CREATE,
			name = "dialogs"
		},
		trailerFillLevel = {
			active = false,
			loadingState = AdditionalSettings.LOADING_STATE.ON_CREATE,
			name = "trailerFillLevel"
		},
		easyMotorStart = {
			active = false,
			loadingState = AdditionalSettings.LOADING_STATE.ON_CREATE,
			name = "easyMotorStart"
		},
		autoStart = {
			active = false,
			loadingState = AdditionalSettings.LOADING_STATE.ON_CREATE,
			name = "autoStart"
		},
		basicMachines = {
			active = true,
			loadingState = AdditionalSettings.LOADING_STATE.ON_CREATE,
			name = "basicMachines"
		},
		dof = {
			active = true,
			target = self,
			callbackFunc = self.changeDofState,
			createFunc = self.createDofSettings,
			loadingState = AdditionalSettings.LOADING_STATE.ON_MISSION_START,
			name = "dof"
		},
		birds = {
			active = true,
			loadingState = AdditionalSettings.LOADING_STATE.ON_CREATE,
			name = "birds"
		},
		animalsCollisions = {
			active = true,
			target = self,
			callbackFunc = self.changeAnimalsCollisionMask,
			loadingState = AdditionalSettings.LOADING_STATE.ON_CREATE,
			name = "animalsCollisions"
		},
		cameraCollisions = {
			active = true,
			loadingState = AdditionalSettings.LOADING_STATE.ON_CREATE,
			name = "cameraCollisions"
		},
		crosshair = {
			state = 0,
			loadingState = AdditionalSettings.LOADING_STATE.ON_MISSION_START,
			name = "crosshair"
		},
		currentDate = {
			state = 0,
			target = self,
			callbackFunc = self.changeCurrentDate,
			createFunc = self.createCurrentDateSettings,
			loadingState = AdditionalSettings.LOADING_STATE.ON_MISSION_START,
			name = "currentDate"
		},
		fadeEffect = {
			state = 0,
			target = self,
			callbackFunc = self.changeFadeEffectDuration,
			createFunc = self.createFadeEffectSettings,
			loadingState = AdditionalSettings.LOADING_STATE.ON_MISSION_START,
			name = "fadeEffect"
		},
		smoothVehicleCamera = {
			state = 0,
			target = self,
			callbackFunc = self.changeVehicleCameraValue,
			createFunc = self.createSmoothVehicleCameraSettings,
			loadingState = AdditionalSettings.LOADING_STATE.ON_MISSION_START,
			name = "smoothVehicleCamera"
		},
		smoothPlayerCamera = {
			state = 0,
			target = self,
			callbackFunc = self.changePlayerCameraValue,
			createFunc = self.createSmoothPlayerCameraSettings,
			loadingState = AdditionalSettings.LOADING_STATE.ON_MISSION_START,
			name = "smoothPlayerCamera"
		},
		lighting = {
			state = 0,
			target = self,
			callbackFunc = self.changeLighting,
			loadingState = AdditionalSettings.LOADING_STATE.ON_MISSION_START,
			name = "lighting"
		},
		brightness = {
			state = tonumber(string.format("%.0f", (getBrightness() - 0.5) / 0.1)),
			target = self,
			callbackFunc = self.changeBrightness,
			loadingState = AdditionalSettings.LOADING_STATE.ON_MISSION_START,
			name = "brightness"
		},
		pauseGame = {
			active = false,
			target = self,
			callbackFunc = self.changeGameState,
			loadingState = AdditionalSettings.LOADING_STATE.NO_LOAD,
			name = "pauseGame"
		}
	}

	for _, setting in pairs(settings) do
		self.callFunction("create", setting)
	end

	return settings
end

function AdditionalSettings.callFunction(funcName, object, ...)
	local func = object[funcName .. "Func"]

	if func ~= nil then
		if object.target ~= nil then
			if object.target ~= nil then
				func(object.target, object, ...)
			else
				func(object, ...)
			end
		end
	end
end

function AdditionalSettings:getValue(setting)
	return self.addittionalSettings[setting].active or self.addittionalSettings[setting].state
end

function AdditionalSettings:getSettings(setting)
	return self.addittionalSettings[setting]
end

function AdditionalSettings:overwriteFunctions(loadingState)
	if loadingState == AdditionalSettings.LOADING_STATE.ON_CREATE then
		local oldCreateAnimalHusbandry = getfenv(0).createAnimalHusbandry
		getfenv(0).createAnimalHusbandry = function(animalType, navMeshNode, animalHusbandryXMLFilename, placementRaycastDistance, collisionMaskFilter)
			if collisionMaskFilter == 4294967295.0 and not g_additionalSettings:getValue("animalsCollisions") then
				collisionMaskFilter = 268435456
			end
			return oldCreateAnimalHusbandry(animalType, navMeshNode, animalHusbandryXMLFilename, placementRaycastDistance, collisionMaskFilter)
		end

		LightWildlife.removeFarAwayAnimals = Utils.overwrittenFunction(LightWildlife.removeFarAwayAnimals, self.lightWildlife_removeFarAwayAnimals)
		self.currentMission.shopMenu.onClickBrand = Utils.overwrittenFunction(self.currentMission.shopMenu.onClickBrand, self.shopMenu_onClickBrand)
		self.currentMission.shopMenu.onClickItemCategory = Utils.overwrittenFunction(self.currentMission.shopMenu.onClickItemCategory, self.shopMenu_onClickItemCategory)
		self.mpLoadingScreen.setStatusText = Utils.overwrittenFunction(self.mpLoadingScreen.setStatusText, self.mpLoadingScreen_setStatusText)
	elseif loadingState == AdditionalSettings.LOADING_STATE.ON_MAP_LOAD then
		local oldSetCamera = getfenv(0).setCamera
		getfenv(0).setCamera = function(camera)
			g_additionalSettings:fadeScreen()
			oldSetCamera(camera)
		end

		self.gui.changeScreen = Utils.overwrittenFunction(self.gui.changeScreen, self.gui_changeScreen)
		self.gui.showInfoDialog = Utils.appendedFunction(self.gui.showInfoDialog, self.gui_showInfoDialog)
		self.gui.showMessageDialog = Utils.appendedFunction(self.gui.showMessageDialog, self.gui_showMessageDialog)
		self.mpLoadingScreen.onReadyToStart = Utils.appendedFunction(self.mpLoadingScreen.onReadyToStart, self.mpLoadingScreen_onReadyToStart)
		self.shopConfigScreen.elements[1].onClose = Utils.appendedFunction(self.shopConfigScreen.elements[1].onClose, self.shopConfigScreen_onClose)
		Player.onInputLookLeftRight = Utils.overwrittenFunction(Player.onInputLookLeftRight, self.player_onInputLookLeftRight)
		Player.onInputLookUpDown = Utils.overwrittenFunction(Player.onInputLookUpDown, self.player_onInputLookUpDown)
		VehicleCamera.actionEventLookUpDown = Utils.overwrittenFunction(VehicleCamera.actionEventLookUpDown, self.vehicleCamera_actionEventLookUpDown)
		VehicleCamera.actionEventLookLeftRight = Utils.overwrittenFunction(VehicleCamera.actionEventLookLeftRight, self.vehicleCamera_actionEventLookLeftRight)

		if not g_modIsLoaded["FS19_disableVehicleCameraCollision"] then
			VehicleCamera.getCollisionDistance = Utils.overwrittenFunction(VehicleCamera.getCollisionDistance, self.vehicleCamera_getCollisionDistance)
		end

		self.currentMission.hud.gameInfoDisplay.drawTimeText = Utils.prependedFunction(self.currentMission.hud.gameInfoDisplay.drawTimeText, self.gameInfoDisplay_drawTimeText)
		self.currentMission.hud.gameInfoDisplay.updateTime = Utils.overwrittenFunction(self.currentMission.hud.gameInfoDisplay.updateTime, self.gameInfoDisplay_updateTime)
	elseif loadingState == AdditionalSettings.LOADING_STATE.ON_MISSION_START then
		self.currentMission.player.pickedUpObjectOverlay.setUVs = Utils.overwrittenFunction(self.currentMission.player.pickedUpObjectOverlay.setUVs, self.pickedUpObjectOverlay_setUVs)
	end
end

function AdditionalSettings:loadSettingsFromXMLFile(directory, loadingState)
	if directory ~= nil then
		local filename = directory .. "/additionalSettings.xml"

		if fileExists(filename) then
			local xmlFile = loadXMLFile("additionalSettings", filename)

			if xmlFile ~= nil then
				for _, setting in pairs(self.addittionalSettings) do
					if setting.loadingState == loadingState then
						if setting.name == "lighting" then
							local lightingName = getXMLString(xmlFile, "additionalSettings.lighting")
							if lightingName ~= nil then
								for i, lighting in pairs(self.customLightings) do
									if lighting.name == lightingName then
										setting.state = i
										self.callFunction("callback", setting, i)
									end
								end
							end
						else
							if setting.active ~= nil then
								local active = getXMLBool(xmlFile, "additionalSettings." .. setting.name)
								if active ~= nil and setting.active ~= active then
									setting.active = active
									self.callFunction("callback", setting, active)
								end
							elseif setting.state ~= nil then
								local state = getXMLInt(xmlFile, "additionalSettings." .. setting.name)
								if state ~= nil and setting.state ~= state then
									setting.state = state
									self.callFunction("callback", setting, state)
								end
							end
						end
					end
				end

				delete(xmlFile)
			end
		end
	end
end

function AdditionalSettings:saveSettingsToXMLFile(directory)
	if directory ~= nil then
		local xmlFile = createXMLFile("additionalSettings", directory .. "/additionalSettings.xml", "additionalSettings")

		if xmlFile ~= nil then
			for _, setting in pairs(self.addittionalSettings) do
				if setting.name == "lighting" then
					if setting.state ~= 0 then
						setXMLString(xmlFile, "additionalSettings.lighting", self.customLightings[setting.state].name)
					end
				else
					if setting.active ~= nil then
						setXMLBool(xmlFile, "additionalSettings." .. setting.name, setting.active)
					elseif setting.state ~= nil then
						setXMLInt(xmlFile, "additionalSettings." .. setting.name, setting.state)
					end
				end
			end

			saveXMLFile(xmlFile)
			delete(xmlFile)
		end
	end
end

function AdditionalSettings:copyModEnvironmentTexts(add)
	local globalTexts = getfenv(0).g_i18n.texts

	for name, text in pairs(self.i18n.texts) do
		if not add then
			text = nil
		end

		globalTexts[name] = text
	end
end

function AdditionalSettings:getCrosshairTexts()
	return {
		self.i18n:getText("ui_on"),
		self.i18n:getText("additionalSettings_ui_crosshairPartial"),
		self.i18n:getText("ui_off")
	}
end

function AdditionalSettings:getCurrentDateTexts()
	return {
		self.i18n:getText("ui_off"),
		string.format("%s", self.i18n:getText("additionalSettings_ui_currentTimeHour")),
		string.format("%s + %s", self.i18n:getText("additionalSettings_ui_currentTimeHour"), self.i18n:getText("additionalSettings_ui_currentTimeDate")),
		string.format("%s + %s (%s)", self.i18n:getText("additionalSettings_ui_currentTimeHour"), self.i18n:getText("additionalSettings_ui_currentTimeDate"), self.i18n:getText("ui_details"))
	}
end

function AdditionalSettings:getSmoothCameraTexts()
	local texts = {self.i18n:getText("ui_off")}

	for i = 10, 100 + 0.0001, 10 do
		table.insert(texts, string.format("%d%%", i))
	end

	return texts
end

function AdditionalSettings:getLightingTexts()
	return self.customLightingsTexts
end

function AdditionalSettings:getBrightnessTexts()
	local texts = {}

	for i = 0.5, 3 + 0.0001, 0.1 do
		table.insert(texts, string.format("%.1f", i))
	end

	return texts
end

function AdditionalSettings:getFadeTexts()
	local texts = {self.i18n:getText("ui_off")}

	for i = 50, 800 + 0.0001, 50 do
		table.insert(texts, string.format("%d ms", i))
	end

	return texts
end

function AdditionalSettings:getSmoothCameraValueFromIndex(index)
	return 5 - index * 0.35
end

function AdditionalSettings:getFadeEffectDurationFromIndex(index)
	return index * 50
end

function AdditionalSettings:getBrightnessFromIndex(index)
	return MathUtil.clamp(0.5 + 0.1 * index, 0.5, 3)
end

function AdditionalSettings:loadMap(filename)
	self:overwriteFunctions(AdditionalSettings.LOADING_STATE.ON_MAP_LOAD)

	local lightingDirectory = g_modsSettingsPath .. "lighting/"
	createFolder(lightingDirectory)
	self:copyBaseLightingFiles(self.baseDirectory .. "xml/lighting/", lightingDirectory .. "sample/")
	self.customLightings, self.customLightingsTexts = self:loadCustomLightingFiles(lightingDirectory, filename)

	self.lightingDirectory = lightingDirectory
	self.currentMapI3dFilename = filename

	local fadeOverlay = Overlay:new(g_baseHUDFilename, 0, 0, 1, 1)
	fadeOverlay:setUVs(getNormalizedUVs(HUD.UV.AREA))
	fadeOverlay:setColor(0, 0, 0, 0)
	self.fadeScreenElement = HUDElement:new(fadeOverlay)
	self.fadeAnimation = TweenSequence.NO_SEQUENCE

	if AdditionalSettings.SHOW_SECONDS then
		self.createClockSecondHand(self.currentMission.hud.gameInfoDisplay)
	end

	local pageAdditionalSettings = self.currentMission.inGameMenu.pageAdditionalSettings
	if pageAdditionalSettings ~= nil then
		pageAdditionalSettings:initialize(self)
	end

	if g_addTestCommands then
		addConsoleCommand("gsReloadCustomLightingSettings", "", "consoleCommandReloadCustomLightingSettings", self)
	end
end

function AdditionalSettings:onMissionStarted()
	self:overwriteFunctions(AdditionalSettings.LOADING_STATE.ON_MISSION_START)
	self:loadSettingsFromXMLFile(g_modsSettingsPath, AdditionalSettings.LOADING_STATE.ON_MISSION_START)
end

function AdditionalSettings:deleteMap()
	self:copyModEnvironmentTexts(false)
	removeConsoleCommand("gsReloadCustomLightingSettings")
	removeModEventListener(self)

	getfenv(0).g_additionalSettings = nil
end

function AdditionalSettings:update(dt)
	if self.currentMission:getIsClient() then
		if not self.fadeAnimation:getFinished() then
			self.fadeAnimation:update(dt)
		end
	end
end

function AdditionalSettings:draw()
	if self.fadeScreenElement:getVisible() then
		self.fadeScreenElement:draw()
	end
end

function AdditionalSettings:copyBaseLightingFiles(directory, targetDirectory)
	local lighting = directory .. "lighting.xml"
	local colorGrading = directory .. "colorGrading.xml"
	local colorGradingNight = directory .. "colorGradingNight.xml"

	local lightingTarget = targetDirectory .. "lighting.xml"
	local colorGradingTarget = targetDirectory .. "colorGrading.xml"
	local colorGradingNightTarget = targetDirectory .. "colorGradingNight.xml"

	createFolder(targetDirectory)
	copyFile(lighting, lightingTarget, false)
	copyFile(colorGrading, colorGradingTarget, false)
	copyFile(colorGradingNight, colorGradingNightTarget, false)
end

function AdditionalSettings:loadCustomLightingFiles(directory, currentMapI3dFilename)
	local customLightings = {}
	local customLightingsTexts = {self.i18n:getText("ui_off")}
	local baseDirectoryBackup = self.currentMission.baseDirectory
	local files = Files:new(directory)

	for _, v in pairs(files.files) do
		if v.isDirectory then
			local baseDirectory = directory .. v.filename .. "/"
			local xmlFilename = baseDirectory .. "lighting.xml"
			local xmlFile = loadXMLFile("lightingXML", xmlFilename)

			if xmlFile ~= nil then
				local customLighting = Lighting:new()
				local name = v.filename
				local text = name

				if self.i18n:hasText("additionalSettings_customLighting_" .. name) then
					text = self.i18n:getText("additionalSettings_customLighting_" .. name)
				end

				self.currentMission.baseDirectory = baseDirectory
				customLighting:load(xmlFile, "lighting")
				customLighting:setEnvironment(self.currentMission.environment)

				table.insert(customLightings, {lighting = customLighting, name = name})
				table.insert(customLightingsTexts, text)

				g_logManager:info("[%s] '%s' lighting configuration loaded.", text, xmlFilename)

				delete(xmlFile)
			end
		end
	end

	for i = 1, self.mapManager:getNumOfMaps() do
		local map = self.mapManager:getMapDataByIndex(i)
		local mapXMLFilename = Utils.getFilename(map.mapXMLFilename, map.baseDirectory)
		local mapXML = loadXMLFile("MapXML", mapXMLFilename)

		if mapXML ~= nil then
			local i3dFilename = Utils.getFilename(getXMLString(mapXML, "map.filename"), map.baseDirectory)

			if i3dFilename ~= currentMapI3dFilename then
				local environmentXMLFilename = Utils.getFilename(getXMLString(mapXML, "map.environment#filename"), map.baseDirectory)

				if environmentXMLFilename == "data/maps/mapDE_environment.xml" and map.title ~= self.i18n:getText("mapDE_title") or environmentXMLFilename == "data/maps/mapUS_environment.xml" and map.title ~= self.i18n:getText("mapUS_title") then
					g_logManager:info("[%s] '%s' has default lighting configuration.", map.title, environmentXMLFilename)
				else
					local xmlFile = loadXMLFile("lightingXML", environmentXMLFilename)

					if xmlFile ~= nil then
						local timeProbesCount = 0
						while true do
							local key = string.format("environment.envMap.timeProbe(%d)", timeProbesCount)
							if not hasXMLProperty(xmlFile, key) or timeProbesCount > 0 then
								break
							end

							timeProbesCount = timeProbesCount + 1
						end

						if timeProbesCount ~= 0 then
							local customLighting = Lighting:new()

							self.currentMission.baseDirectory = map.baseDirectory
							customLighting:load(xmlFile, "environment")
							customLighting:setEnvironment(self.currentMission.environment)

							table.insert(customLightings, {lighting = customLighting, name = map.title})
							table.insert(customLightingsTexts, map.title)

							g_logManager:info("[%s] '%s' lighting configuration loaded.", map.title, environmentXMLFilename)
						else
							g_logManager:warning("[%s] '%s' lighting configuration has no envMapTimes!", map.title, environmentXMLFilename)
						end

						delete(xmlFile)
					end
				end
			end

			delete(mapXML)
		end
	end

	self.currentMission.baseDirectory = baseDirectoryBackup

	return customLightings, customLightingsTexts
end

function AdditionalSettings:consoleCommandReloadCustomLightingSettings()
	self:changeLighting()
	self:getSettings("lighting").state = 0

	self.currentMission.xmlFile = loadXMLFileFromMemory("fakeXMLFile", "<map/>")
	self.customLightings, self.customLightingsTexts = self:loadCustomLightingFiles(self.lightingDirectory, self.currentMapI3dFilename)

	delete(self.currentMission.xmlFile)
	self.currentMission.xmlFile = nil

	self.currentMission.inGameMenu.pageAdditionalSettings.multiLighting:setTexts(self.customLightingsTexts)

	return string.format("Custom lighting settings updated, available lighting configurations: '%d'", #self.customLightings)
end

function AdditionalSettings:fadeScreen()
	local fadeSettings = self:getSettings("fadeEffect")

	if not self.currentMission.hud:getIsFading() and fadeSettings.effectDuration ~= 0 then
		local seq = TweenSequence.new(self.fadeScreenElement)

		seq:addTween(Tween:new(self.fadeScreenElement.setAlpha, 1, 0, fadeSettings.effectDuration))
		seq:start()

		self.fadeAnimation = seq
	end
end

function AdditionalSettings:updateCurrentDateFormat(state, hourFormat)
	local hourF = "%H"
	local dateF = "%d-%m-%Y"
	local am_pm = ""

	if not hourFormat then
		hourF = "%I"
		dateF = "%m/%d/%Y"
		am_pm = " %p"
	end

	local dateFormat = hourF .. ":%M:%S" .. am_pm
	if state == 2 then
		dateFormat = dateFormat .. " | " .. dateF
	elseif state == 3 then
		dateFormat = dateFormat .. " | %A, %d %B %Y"
	end

	self:getSettings("currentDate").dateFormat = dateFormat
end

function AdditionalSettings.createClockSecondHand(object)
	local posX, posY = object.clockHandLarge:getPosition()
	local size = {4.5, 22}
	local uvs = GameInfoDisplay.UV.CLOCK_HAND_LARGE
	local color = {1, 0, 0, 0.7}
	local pivot = GameInfoDisplay.PIVOT.CLOCK_HAND_LARGE
	local pivotX, pivotY = object:normalizeUVPivot(pivot, size, uvs)

	object.clockSecondHand = object:createClockHand(g_baseHUDFilename, posX + pivotX, posY + pivotY, size, uvs, color, pivot)
	object.clockElement:addChild(object.clockSecondHand)
end

function AdditionalSettings:createDofSettings(setting)
	setting.initialStateBackup = self.depthOfFieldManager.initialState
end

function AdditionalSettings:createDialogsSettings(setting)
	setting.infoDialogs = {
		{text = "shop_messagePurchaseReady", callback = true},
		{text = "shop_messageBoughtChainsaw", callback = true},
		{text = "shop_messageLeasingReady", callback = true},
		{text = "ui_vehicleResetDone", callback = true},
		{text = "shop_messageReturnedVehicle", callback = true}
	}

	setting.messageDialogs = {
		"shop_messageBuyingVehicle",
		"shop_messageLeasingVehicle"
	}
end

function AdditionalSettings:createCurrentDateSettings(setting)
	setting.dateFormat = ""

	setting.dayName = {
		["Monday"] = self.i18n:getText("ui_financesDay1"),
		["Tuesday"] = self.i18n:getText("ui_financesDay2"),
		["Wednesday"] = self.i18n:getText("ui_financesDay3"),
		["Thursday"] = self.i18n:getText("ui_financesDay4"),
		["Friday"] = self.i18n:getText("ui_financesDay5"),
		["Saturday"] = self.i18n:getText("ui_financesDay6"),
		["Sunday"] = self.i18n:getText("ui_financesDay7")
	}

	setting.monthName = {
		["January"] = self.i18n:getText("additionalSettings_ui_monthJanuary"),
		["February"] = self.i18n:getText("additionalSettings_ui_monthFebruary"),
		["March"] = self.i18n:getText("additionalSettings_ui_monthMarch"),
		["April"] = self.i18n:getText("additionalSettings_ui_monthApril"),
		["May"] = self.i18n:getText("additionalSettings_ui_monthMay"),
		["June"] = self.i18n:getText("additionalSettings_ui_monthJune"),
		["July"] = self.i18n:getText("additionalSettings_ui_monthJuly"),
		["August"] = self.i18n:getText("additionalSettings_ui_monthAugust"),
		["September"] = self.i18n:getText("additionalSettings_ui_monthSeptember"),
		["October"] = self.i18n:getText("additionalSettings_ui_monthOctober"),
		["November"] = self.i18n:getText("additionalSettings_ui_monthNovember"),
		["December"] = self.i18n:getText("additionalSettings_ui_monthDecember")
	}
end

function AdditionalSettings:createSmoothVehicleCameraSettings(setting)
	setting.yaw = 0
	setting.pitch = 0
	setting.value = 0
end

function AdditionalSettings:createSmoothPlayerCameraSettings(setting)
	setting.yaw = 0
	setting.pitch = 0
	setting.value = 0
end

function AdditionalSettings:createFadeEffectSettings(setting)
	setting.effectDuration = 0
end

function AdditionalSettings:changeHudVisibility(setting, state, uiElement)
	self.currentMission.hud:setIsVisible(state)
end

function AdditionalSettings:changeHourFormat(setting, state, uiElement)
	local gameInfoDisplay = self.currentMission.hud.gameInfoDisplay
	local timeBoxWidth = GameInfoDisplay.SIZE.TIME_BOX[1]

	if not state then
		timeBoxWidth = GameInfoDisplay.SIZE.TIME_BOX[1] + 37.5
	end

	local boxWidth, boxHeight = gameInfoDisplay:scalePixelToScreenVector({timeBoxWidth, 0})

	local sepX, sepY = gameInfoDisplay.timeBox.separator:getPosition()
	local arrowX, arrowY = gameInfoDisplay.timeScaleArrow:getPosition()
	local fArrowX, fArrowY = gameInfoDisplay.timeScaleArrowFast:getPosition()

	gameInfoDisplay.timeBox.overlay.width = boxWidth

	gameInfoDisplay:storeScaledValues()
	gameInfoDisplay:updateSizeAndPositions()

	gameInfoDisplay.timeBox.separator:setPosition(sepX, sepY)
	gameInfoDisplay.timeScaleArrow:setPosition(arrowX, arrowY)
	gameInfoDisplay.timeScaleArrowFast:setPosition(fArrowX, fArrowY)

	self:updateCurrentDateFormat(self:getValue("currentDate"), state)
end

function AdditionalSettings:changeAnimalsCollisionMask(setting, state, uiElement)
	if self.gui.currentGuiName == "InGameMenu" then
		self.gui:showInfoDialog({
			text = self.i18n:getText("additionalSettings_dialog_animalsCollisions"),
			dialogType = DialogElement.TYPE_INFO
		})
	end

	if uiElement ~= nil then
		local titleElement = uiElement.elements[4]
		local text = string.format("%s*", self.i18n:getText("additionalSettings_animalsCollisions"))

		if titleElement.text ~= text then
			titleElement:setText(text)
		end
	end
end

function AdditionalSettings:changeDofState(setting, state, uiElement)
	local initialState = {0, 0, 0, math.huge, math.huge}

	if state then
		initialState = setting.initialStateBackup
	end

	self.depthOfFieldManager.initialState = initialState
	self.depthOfFieldManager:reset()
end

function AdditionalSettings:changeBrightness(setting, state, uiElement)
	setSceneBrightness(self:getBrightnessFromIndex(state))
end

function AdditionalSettings:changeLighting(setting, state, uiElement)
	local lighting = nil

	if state ~= nil and state ~= 0 then
		lighting = self.customLightings[state].lighting
	end

	self.currentMission.environment:setCustomLighting(lighting)
end

function AdditionalSettings:changeVehicleCameraValue(setting, state, uiElement)
	setting.value = self:getSmoothCameraValueFromIndex(state)
end

function AdditionalSettings:changePlayerCameraValue(setting, state, uiElement)
	setting.value = self:getSmoothCameraValueFromIndex(state)
end

function AdditionalSettings:changeFadeEffectDuration(setting, state, uiElement)
	setting.effectDuration = self:getFadeEffectDurationFromIndex(state)
end

function AdditionalSettings:changeCurrentDate(setting, state, uiElement)
	self:updateCurrentDateFormat(state, self:getValue("hourFormat"))
end

function AdditionalSettings:changeGameState(setting, state, uiElement)
end

function AdditionalSettings:updateHudVisibility(setting, uiElement)
	setting.active = self.currentMission.hud:getIsVisible()
end

function AdditionalSettings.gui_changeScreen(object, superFunc, source, screenClass, returnScreenClass)
	local disableFadeEffect = object.currentGuiName == "ChatDialog"
	local retValue = superFunc(object, source, screenClass, returnScreenClass)

	if not disableFadeEffect and object.currentGuiName == "" then
		g_additionalSettings:fadeScreen()
	end

	return retValue
end

function AdditionalSettings.gui_showInfoDialog(object, args)
	if not g_additionalSettings:getValue("dialogs") then
		local dialog = object.guis["InfoDialog"]

		if dialog ~= nil and args ~= nil then
			for _, infoDialog in pairs(g_additionalSettings:getSettings("dialogs").infoDialogs) do
				if args.text == g_i18n:getText(infoDialog.text) then
					dialog.target:close()

					if infoDialog.callback and args.callback ~= nil then
						if args.target ~= nil then
							args.callback(args.target, args.args)
						else
							args.callback(args.args)
						end
					end
				end
			end
		end
	end
end

function AdditionalSettings.gui_showMessageDialog(object, args)
	if not g_additionalSettings:getValue("dialogs") then
		local dialog = object.guis["MessageDialog"]

		if dialog ~= nil and args ~= nil then
			for _, dialogText in pairs(g_additionalSettings:getSettings("dialogs").messageDialogs) do
				if args.text == g_i18n:getText(dialogText) then
					dialog.target:close()
				end
			end
		end
	end
end

function AdditionalSettings.lightWildlife_removeFarAwayAnimals(object, superFunc, maxDistance, refPosX, refPosY, refPosZ)
	if not g_additionalSettings:getValue("birds") then
		maxDistance = 0
	end

	return superFunc(object, maxDistance, refPosX, refPosY, refPosZ)
end

function AdditionalSettings.shopMenu_onClickBrand(object, superFunc, brandId, _, _, categoryDisplayName)
	local brandItems = object.shopController:getItemsByBrand(brandId)

	object.currentDisplayItems = brandItems

	if not g_additionalSettings:getValue("basicMachines") then
		object.currentDisplayItems = {}
	
		for i = 1, #brandItems do
			if brandItems[i].storeItem.customEnvironment ~= nil then
				table.insert(object.currentDisplayItems, brandItems[i])
			end
		end
	end

	object.pageShopItemDetails:setDisplayItems(object.currentDisplayItems, false)
	object.pageShopItemDetails:setCategory(getNormalizedUVs(ShopMenu.TAB_UV.BRANDS), object.l10n:getText(ShopMenu.L10N_SYMBOL.HEADER_BRANDS), categoryDisplayName)
	object:pushDetail(object.pageShopItemDetails)
end

function AdditionalSettings.shopMenu_onClickItemCategory(object, superFunc, categoryName, baseCategoryIconUVs, baseCategoryDisplayName, categoryDisplayName)
	local categoryItems = object.shopController:getItemsByCategory(categoryName)
	
	object.currentDisplayItems = categoryItems

	if not g_additionalSettings:getValue("basicMachines") then
		object.currentDisplayItems = {}
	
		for i = 1, #categoryItems do
			if categoryItems[i].storeItem.customEnvironment ~= nil then
				table.insert(object.currentDisplayItems, categoryItems[i])
			end
		end
	end

	object.pageShopItemDetails:setDisplayItems(object.currentDisplayItems, false)
	object.pageShopItemDetails:setCategory(baseCategoryIconUVs, baseCategoryDisplayName, categoryDisplayName)
	object:pushDetail(object.pageShopItemDetails)
end

function AdditionalSettings.mpLoadingScreen_onReadyToStart(object)
	if g_additionalSettings:getValue("autoStart") and g_currentMission:canStartMission() and not g_gui:getIsDialogVisible() then
		object:onClickOk()
	end
end

function AdditionalSettings.mpLoadingScreen_setStatusText(object, superFunc, text)
	if g_additionalSettings:getValue("autoStart") and text == g_i18n:getText("ui_gameIsLoadingPleaseWait") then
		text = string.format("%s (%s)", text, g_i18n:getText("additionalSettings_autoStart"))
	end

	superFunc(object, text)
end

function AdditionalSettings.shopConfigScreen_onClose(object, element)
	g_additionalSettings:changeBrightness(nil, g_additionalSettings:getValue("brightness"), nil)
	g_additionalSettings:changeLighting(nil, g_additionalSettings:getValue("lighting"), nil)
end

function AdditionalSettings.player_onInputLookLeftRight(object, superFunc, actionName, inputValue, callbackState, isAnalog, isMouse)
	if g_additionalSettings:getValue("smoothPlayerCamera") ~= 0 then
		if not object.lockedInput then
			if isMouse then
				inputValue = inputValue * 0.001 * 16.666
			else
				inputValue = inputValue * g_currentDt *0.001
			end

			local playerCameraSetings = g_additionalSettings:getSettings("smoothPlayerCamera")

			playerCameraSetings.yaw = inputValue + math.pow(0.99579, g_currentDt * playerCameraSetings.value) * (playerCameraSetings.yaw - inputValue)
			object.inputInformation.yawCamera = object.inputInformation.yawCamera + playerCameraSetings.yaw 
		end

		object.inputInformation.isMouseRotation = isMouse
	else
		superFunc(object, actionName, inputValue, callbackState, isAnalog, isMouse)
	end
end

function AdditionalSettings.player_onInputLookUpDown(object, superFunc, actionName, inputValue, callbackState, isAnalog, isMouse)
	if g_additionalSettings:getValue("smoothPlayerCamera") ~= 0 then
		if not object.lockedInput then
			local pitchValue = g_gameSettings:getValue("invertYLook") and -inputValue or inputValue

			if isMouse then
				pitchValue = pitchValue * 0.001 * 16.666
			else
				pitchValue = pitchValue * g_currentDt * 0.001
			end

			local playerCameraSetings = g_additionalSettings:getSettings("smoothPlayerCamera")

			playerCameraSetings.pitch = pitchValue + math.pow(0.99579, g_currentDt * playerCameraSetings.value) * (playerCameraSetings.pitch - pitchValue)
			object.inputInformation.pitchCamera = object.inputInformation.pitchCamera + playerCameraSetings.pitch
		end
	else
		superFunc(object, actionName, inputValue, callbackState, isAnalog, isMouse)
	end
end

function AdditionalSettings.vehicleCamera_actionEventLookLeftRight(object, superFunc, actionName, inputValue, callbackState, isAnalog, isMouse)
	if g_additionalSettings:getValue("smoothVehicleCamera") ~= 0 then
		if isMouse then
			inputValue = inputValue * 0.001 * 16.666
		else
			inputValue = inputValue * 0.001 * g_currentDt
		end

		local vehicleCameraSettings = g_additionalSettings:getSettings("smoothVehicleCamera")

		vehicleCameraSettings.yaw = inputValue + math.pow(0.99579, g_currentDt * vehicleCameraSettings.value) * (vehicleCameraSettings.yaw - inputValue)
		object.lastInputValues.leftRight = object.lastInputValues.leftRight + vehicleCameraSettings.yaw
	else
		superFunc(object, actionName, inputValue, callbackState, isAnalog, isMouse)
	end
end

function AdditionalSettings.vehicleCamera_actionEventLookUpDown(object, superFunc, actionName, inputValue, callbackState, isAnalog, isMouse)
	if g_additionalSettings:getValue("smoothVehicleCamera") ~= 0 then
		if isMouse then
			inputValue = inputValue * 0.001 * 16.666
		else
			inputValue = inputValue * 0.001 * g_currentDt
		end

		local vehicleCameraSettings = g_additionalSettings:getSettings("smoothVehicleCamera")

		vehicleCameraSettings.pitch = inputValue + math.pow(0.99579, g_currentDt * vehicleCameraSettings.value) * (vehicleCameraSettings.pitch - inputValue)
		object.lastInputValues.upDown = object.lastInputValues.upDown + vehicleCameraSettings.pitch
	else
		superFunc(object, actionName, inputValue, callbackState, isAnalog, isMouse)
	end
end

function AdditionalSettings.vehicleCamera_getCollisionDistance(object, superFunc)
	if not g_additionalSettings:getValue("cameraCollisions") then
		return false, nil, nil, nil, nil, nil
	end

	return superFunc(object)
end

function AdditionalSettings.gameInfoDisplay_updateTime(object, superFunc)
	local currentTime = object.environment.dayTime / 3600000
	local timeHours, timeMinutes = math.modf(currentTime)
	local timeMinutes, timeSeconds = math.modf(timeMinutes * 60)
	timeSeconds = math.floor(timeSeconds * 60)

	local timeHoursText = timeHours
	local am_pm = ""

	if not g_additionalSettings:getValue("hourFormat") then
		am_pm = " AM"

		if timeHoursText == 0 then
			timeHoursText = 12
		elseif timeHoursText > 11 then
			if timeHoursText > 12 then
				timeHoursText = timeHoursText - 12
			end

			am_pm = " PM"
		end
	end

	local timeTextFormat = string.format("%02d:%02d%s", timeHoursText, timeMinutes, am_pm)

	if AdditionalSettings.SHOW_SECONDS then
		timeTextFormat = string.format("%02d:%02d:%02d%s", timeHoursText, timeMinutes, timeSeconds, am_pm)

		if object.clockSecondHand ~= nil then
			local secondsRotation = -(timeSeconds/60) * math.pi * 2
			object.clockSecondHand:setRotation(secondsRotation)
		end
	end

	object.timeText = timeTextFormat
	object.timeScaleText = string.format("%d", object.missionInfo.timeScale)

	local hourRotation = -(currentTime % 12 / 12) * math.pi * 2
	local minutesRotation = -(currentTime - timeHours) * math.pi * 2

	object.clockHandSmall:setRotation(hourRotation)
	object.clockHandLarge:setRotation(minutesRotation)

	local isTimeScaleFast = object.missionInfo.timeScale > 1
	object.timeScaleArrow:setVisible(not isTimeScaleFast)
	object.timeScaleArrowFast:setVisible(isTimeScaleFast)
end

function AdditionalSettings.gameInfoDisplay_drawTimeText(object)
	local currentDateState = g_additionalSettings:getValue("currentDate")

	if currentDateState ~= 0 then
		setTextBold(false)
		setTextAlignment(RenderText.ALIGN_RIGHT)
		setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_TOP)
		setTextColor(unpack(GameInfoDisplay.COLOR.TEXT))

		local dateSettings = g_additionalSettings:getSettings("currentDate")
		local dateFormat = getDate(dateSettings.dateFormat)

		if currentDateState == 3 and g_languageShort ~= "en" then
			local dayName = (getDate("%A"))
			local dayFormat = dateSettings.dayName[dayName]

			local monthName = (getDate("%B"))
			local monthFormat = dateSettings.monthName[monthName]

			if dayFormat ~= nil and monthFormat ~= nil then
				dateFormat = string.gsub(dateFormat, dayName, dayFormat)
				dateFormat = string.gsub(dateFormat, monthName, monthFormat)
			end
		end

		renderText(object.moneyTextPositionX, 0.999, object.moneyTextSize * 0.8, dateFormat)
		setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_BASELINE)
		setTextAlignment(RenderText.ALIGN_LEFT)
	end
end

function AdditionalSettings.pickedUpObjectOverlay_setUVs(object, superFunc, uvs)
	local crosshairState = g_additionalSettings:getValue("crosshair")

	if crosshairState ~= 0 then
		if crosshairState == 1 then
			if uvs == g_additionalSettings.currentMission.player.pickedUpObjectAimingUVs then
				uvs = {0, 0, 0, 0, 0, 0, 0, 0}
			end
		else
			uvs = {0, 0, 0, 0, 0, 0, 0, 0}
		end
	end

	superFunc(object, uvs)
end