FillIndicator = {}

function FillIndicator.prerequisitesPresent(specializations)
	return true
end

function FillIndicator.registerEventListeners(vehicleType)
	SpecializationUtil.registerEventListener(vehicleType, "onLoad", FillIndicator)
	SpecializationUtil.registerEventListener(vehicleType, "onUpdate", FillIndicator)
	SpecializationUtil.registerEventListener(vehicleType, "onFillUnitFillLevelChanged", FillIndicator)
end

function FillIndicator:onLoad(savegame)
	local spec = self.spec_fillIndicator

	spec.fillIndicatorActive = false
	spec.fillIndicatorEnabled = 0
	spec.fillIndicatorColor = {1, 1, 1, 1}
	spec.fadeDuration = 500
	spec.startFadeEnabled = 0
	spec.endFadeEnabled = 0
	spec.maxDrawDistance = 500
end

function FillIndicator:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	if g_additionalSettings:getValue("trailerFillLevel") then
		local spec = self.spec_fillIndicator

		spec.fillIndicatorEnabled = math.max(spec.fillIndicatorEnabled - dt, 0)

		local fillIndicatorActive = spec.fillIndicatorEnabled > 0
		spec.fillIndicatorColor[4] = 1

		if spec.fillIndicatorActive ~= fillIndicatorActive then
			if fillIndicatorActive then
				spec.startFadeEnabled = 0
			else
				spec.endFadeEnabled = spec.fadeDuration
			end

			spec.fillIndicatorActive = fillIndicatorActive
		end

		spec.startFadeEnabled = math.min(spec.startFadeEnabled + dt, spec.fadeDuration)
		spec.endFadeEnabled = math.max(spec.endFadeEnabled - dt, 0)

		local startFadeActive = spec.startFadeEnabled < spec.fadeDuration
		local endFadeActive = spec.endFadeEnabled > 0

		if startFadeActive then
			spec.fillIndicatorColor[4] = spec.startFadeEnabled / spec.fadeDuration
		elseif endFadeActive then
			spec.fillIndicatorColor[4] = spec.endFadeEnabled / spec.fadeDuration
		end

		if spec.fillIndicatorActive or endFadeActive then
			if not isActiveForInput then
				local x, y, z = getWorldTranslation(self.rootNode)
				local x1, y1, z1 = getWorldTranslation(getCamera())
				local distSq = MathUtil.vector3LengthSq(x - x1, y - y1, z - z1)

				if distSq <= spec.maxDrawDistance then
					local fillType = self:getFillUnitFillType(spec.lastFillUnitIndex)

					if fillType ~= nil and fillType ~= FillType.UNKNOWN then
						local fillLevel = self:getFillUnitFillLevel(spec.lastFillUnitIndex)
						local capacity = self:getFillUnitCapacity(spec.lastFillUnitIndex)
						local value = self:getFillUnitFillLevelPercentage(spec.lastFillUnitIndex)
						local fillText = string.format("%s - %d/%d (%d%%)", g_fillTypeManager:getFillTypeByIndex(fillType).title, MathUtil.round(fillLevel), MathUtil.round(capacity), 100 * value)

						Utils.renderTextAtWorldPosition(x, y + 3, z, fillText, getCorrectTextSize(0.02), 0, spec.fillIndicatorColor)
					end
				end
			end

			self:raiseActive()
		end
	end
end

function FillIndicator:onFillUnitFillLevelChanged(fillUnitIndex, fillLevelDelta, fillTypeIndex, toolType, fillPositionData, appliedDelta)
	if g_additionalSettings:getValue("trailerFillLevel") then
		local spec = self.spec_fillIndicator

		spec.fillIndicatorEnabled = 1500
		spec.lastFillUnitIndex = fillUnitIndex
	end
end